#include <stdio.h>
#include <stdlib.h>
#include <cuda.h>
#include "customer_functions.h"
#include "cudaEGL.h"
#include "iva_metadata.h"
#include "opencv2/core.hpp"
#include "opencv2/calib3d.hpp"
#include "opencv2/cudawarping.hpp" 
#include <opencv2/cudaarithm.hpp>

const int max_width = 3840;
const int max_height = 2160;

static cv::cuda::GpuMat gpu_xmap, gpu_ymap;
cv::cuda::Stream stream[1];

static void pre_process (void **sBaseAddr,unsigned int *smemsize,unsigned int *swidth,unsigned int *sheight,unsigned int *spitch,ColorFormat  *sformat,unsigned int nsurfcount, void ** usrptr){}

static void post_process (void **sBaseAddr,unsigned int *smemsize,unsigned int *swidth,unsigned int *sheight,unsigned int *spitch,ColorFormat  *sformat,unsigned int nsurfcount,void ** usrptr){}

/**
  * 魚眼レンズ補正を行う
  *
 */
static void cv_process_RGBA(void *pdata, int32_t width, int32_t height)
{
    cv::cuda::GpuMat d_Mat_RGBA(height, width, CV_8UC4, pdata);
    cv::cuda::GpuMat d_Mat_RGBA_Src;
    
    d_Mat_RGBA.copyTo(d_Mat_RGBA_Src,stream[0]);
    
    cv::cuda::remap(d_Mat_RGBA_Src, d_Mat_RGBA, gpu_xmap, gpu_ymap, cv::INTER_NEAREST, cv::BORDER_CONSTANT, cv::Scalar(0.f, 0.f, 0.f, 0.f),stream[0]);
}

/**
  * Performs CUDA Operations on egl image.
  * @param image : EGL image
  */
static void gpu_process (EGLImageKHR image, void ** usrptr)
{
  CUresult status;
  CUeglFrame eglFrame;
  CUgraphicsResource pResource = NULL;

  cudaFree(0);
  status = cuGraphicsEGLRegisterImage(&pResource, image, CU_GRAPHICS_MAP_RESOURCE_FLAGS_NONE);
  if (status != CUDA_SUCCESS) {
    printf("cuGraphicsEGLRegisterImage failed : %d \n", status);
    return;
  }

  status = cuGraphicsResourceGetMappedEglFrame( &eglFrame, pResource, 0, 0);
  if (status != CUDA_SUCCESS) {
    printf ("cuGraphicsSubResourceGetMappedArray failed\n");
  }

  //printf ("魚眼補正処理を開始します！ \n");

  if (eglFrame.frameType == CU_EGL_FRAME_TYPE_PITCH) {
    if (eglFrame.eglColorFormat == CU_EGL_COLOR_FORMAT_ABGR) {
 	cv_process_RGBA(eglFrame.frame.pPitch[0], eglFrame.width, eglFrame.height);
    } else
        printf ("Invalid eglcolorformat %d\n", eglFrame.eglColorFormat);
  }

  status = cuGraphicsUnregisterResource(pResource);
  if (status != CUDA_SUCCESS) {
    printf("cuGraphicsEGLUnRegisterResource failed: %d \n", status);
  }
}

extern "C" void
init (CustomerFunction * pFuncs)
{
  printf ("魚眼補正エレメントをロードしました！ \n");

  pFuncs->fPreProcess = pre_process;
  pFuncs->fGPUProcess = gpu_process;
  pFuncs->fPostProcess = post_process;

  cv::Mat xmap(max_height, max_width, CV_32FC1);
  cv::Mat ymap(max_height, max_width, CV_32FC1);
  cv::Mat cam(3, 3, cv::DataType<float>::type);
  cv::Mat dist(4, 1, cv::DataType<float>::type);
  
  cam.at<float>(0, 0) = 1569.08571848262500f;
  cam.at<float>(0, 1) = 0.0f;
  cam.at<float>(0, 2) = 1897.44471769631582f;
  cam.at<float>(1, 0) = 0.0f;
  cam.at<float>(1, 1) = 1590.39243583626353f;
  cam.at<float>(1, 2) = 1063.73079452522120f;
  cam.at<float>(2, 0) = 0.0f;
  cam.at<float>(2, 1) = 0.0f;
  cam.at<float>(2, 2) = 1.0f;
  
  dist.at<float>(0, 0) = -0.20963013646547f;
  dist.at<float>(1, 0) = 0.03491832650144f;
  dist.at<float>(2, 0) = 0.00160953572073f;
  dist.at<float>(3, 0) = -0.00896893124032f;
  dist.at<float>(4, 0) = -0.00194072642273f;
  
  cv::fisheye::initUndistortRectifyMap(cam, dist, cv::Mat(), cam, cv::Size(max_width, max_height), CV_32FC1, xmap, ymap);
	 
  /* upload to GpuMats */
  gpu_xmap.upload(xmap);
  gpu_ymap.upload(ymap);
}

extern "C" void
deinit (void)
{
  /* deinitialization */
}

int main(int argc, char* argv[]) {}